#include "projet.h"

#include <Inventor/Qt/SoQt.h>
#include <q3filedialog.h>
#include <qdir.h>
#include <qmessagebox.h>

#include "form1.h"

#include <Inventor/Qt/viewers/SoQtViewer.h>

#include <math.h>
#include <stdlib.h>

#include <qlabel.h>
#include <qlineedit.h>
#include <qslider.h>
#include <qlcdnumber.h>

// pour la compatibilit MS/Linux de la fonction sleep....
# ifdef _MSC_VER       // si on a un compilateur MS
#include <Windows.h>
# else                 // sinon, Linux/MacOS/BSD
#include <time.h> 
# endif
#include <ctime>
#include <iostream>


extern Form1 *form1;

extern IVLoader *iv;

void millisleep( unsigned int milliseconds )
{
# ifdef _MSC_VER   // si MS
  ::Sleep( milliseconds);
#else  // sinon, sans doute linux/MacOS/BSD
  timespec ts;
  ts.tv_sec = milliseconds / 1000;
  ts.tv_nsec = (milliseconds - ts.tv_sec*1000) * 1000000;
  ::nanosleep(&ts, NULL);
# endif
}

void IVLoader::checkBox1_state_changed(int state)
{
  if (state)
  {
    viewer->setDrawStyle(SoQtViewer::INTERACTIVE, SoQtViewer::VIEW_LINE);
    viewer->setDrawStyle(SoQtViewer::STILL, SoQtViewer::VIEW_LINE);
  }else{
    viewer->setDrawStyle(SoQtViewer::INTERACTIVE, SoQtViewer::VIEW_AS_IS);
    viewer->setDrawStyle(SoQtViewer::STILL, SoQtViewer::VIEW_AS_IS);
  }
}

void IVLoader::center_scene()
{
  viewer->viewAll();
}

void IVLoader::toggle_moving_mode(int state)
{
  mode=1;
}

void IVLoader::toggle_moving_mode2(int state)
{
  mode=0;
}

void IVLoader::toggle_headlight(int state)
{
  if (state)
  {
    viewer->setHeadlight(true);
  }else{
    viewer->setHeadlight(false);
  }
}

void IVLoader::move_plateau(int val)
{
  double bougadou=form1->slider1->value();
  bougadou = -bougadou / 100;
  bougadou = (bougadou)*4+2;
  plateau_translation->translation= SbVec3f(0, bougadou, 0);;
  form1->lCDNumber1->display(bougadou);
}

IVLoader::IVLoader(QWidget* parent) : QObject(parent)
{
  SoQt::init(parent);
  viewer = new SoQtExaminerViewer(parent);
  separator = new SoSeparator;
  viewer->setSceneGraph(separator);
  viewer->setDecoration(false);
  viewer->show();
}

void IVLoader::x_changed(const QString& text)
{
  x_wanted = strtod((char *)text.data(), (char**)NULL);
}

void IVLoader::y_changed(const QString& text)
{
  y_wanted = strtod((char *)text.data(), (char**)NULL);
}

void IVLoader::z_changed(const QString& text)
{
  z_wanted = strtod((char *)text.data(), (char**)NULL);
}


void move_progressif(double x_initial, double x_wanted, double y_initial, double y_wanted, double z_initial, double z_wanted, IVLoader *ivv)
{
  ivv->last_good_x = ivv->x_initial;
  ivv->last_good_y = ivv->y_initial;
  ivv->last_good_z = ivv->z_initial;
  for (double i=0; i<50; i++) {
    if (ivv->gotoposition(x_initial+(x_wanted-x_initial)*(i/50.0f), y_initial+(y_wanted-y_initial)*(i/50.0f), z_initial+(z_wanted-z_initial)*(i/50.0f)))
    {
      ivv->last_good_x = x_initial+(x_wanted-x_initial)*(i/50.0f);
      ivv->last_good_y = y_initial+(y_wanted-y_initial)*(i/50.0f);
      ivv->last_good_z = z_initial+(z_wanted-z_initial)*(i/50.0f);
    } else {
      break;
    }
    ivv->viewer->render();
    millisleep(20); 
  }
  ivv->x_initial = ivv->last_good_x;
  ivv->y_initial = ivv->last_good_y;
  ivv->z_initial = ivv->last_good_z;
}

void IVLoader::toggle_simulation_mode(int state)
{
  simulation_mode = state;
  if (state) {
  }
}


void move_progressif_both(double x_initial, double x_wanted, double y_initial, double y_wanted, double z_initial, double z_wanted, IVLoader *ivv)
{
  for (double i=0; i<70; i++) {
    ivv->item_translation->translation= SbVec3f(x_initial+(x_wanted-x_initial)*(i/70.0f), y_initial+(y_wanted-y_initial)*(i/70.0f)-0.05, z_initial+(z_wanted-z_initial)*(i/70.0f));
    ivv->gotoposition(ivv->x_initial+(x_wanted-ivv->x_initial)*(i/70.0f), ivv->y_initial+(y_wanted-ivv->y_initial)*(i/70.0f), ivv->z_initial+(z_wanted-ivv->z_initial)*(i/70.0f));
    ivv->viewer->render();
    millisleep(20); 
  }
  ivv->x_initial = x_wanted;
  ivv->y_initial = y_wanted;
  ivv->z_initial = z_wanted;
}


void move_progressif_item(double x_initial, double x_wanted, double y_initial, double y_wanted, double z_initial, double z_wanted, IVLoader *ivv)
{
  for (double i=0; i<170; i++) {
    ivv->item_translation->translation= SbVec3f(x_initial+(x_wanted-x_initial)*(i/170.0f), y_initial+(y_wanted-y_initial)*(i/170.0f)-0.05, z_initial+(z_wanted-z_initial)*(i/170.0f));
    if (i>=150){
      ivv->gotoposition(ivv->x_initial+(x_wanted-ivv->x_initial)*((i-150)/20.0f), ivv->y_initial+(y_wanted-ivv->y_initial)*((i-150)/20.0f), ivv->z_initial+(z_wanted-ivv->z_initial)*((i-150)/20.0f));
    }
    ivv->viewer->render();
    millisleep(20); 
  }
  ivv->x_initial = x_wanted;
  ivv->y_initial = y_wanted;
  ivv->z_initial = z_wanted;
}

void move_progressif_item_only(double x_initial, double x_wanted, double y_initial, double y_wanted, double z_initial, double z_wanted, IVLoader *ivv)
{
  for (double i=0; i<30; i++) {
    ivv->item_translation->translation= SbVec3f(x_initial+(x_wanted-x_initial)*(i/30.0f), y_initial+(y_wanted-y_initial)*(i/30.0f)-0.05, z_initial+(z_wanted-z_initial)*(i/30.0f));
    ivv->viewer->render();
    millisleep(20); 
  }
}

void IVLoader::lauch_simulation()
{
  //if (simulation_mode) {
  // on est en mode simulation
  double random = ((double)rand())/(2*(double)RAND_MAX)-0.25;
  item_translation->translation= SbVec3f(random, -0.6, -2);    
  move_progressif_item(-2, -0.3, -0.6, -0.6, random, random, iv);
  move_progressif_both(x_initial, 0, y_initial, -0.5, z_initial, 0.3, iv);
  move_progressif_both(0, 0, -0.5, -0.7, 0.3, 0.3, iv);
  move_progressif_item_only(0, 0, -0.7, -1, 0.3, 0.3, iv);
  //} else {
  // on n'est pas en mode simulation
  //}
}


void IVLoader::move()
{
  if (mode == 0) {
    // on est en mode direct.
    if (gotoposition(x_wanted, y_wanted, z_wanted)) {
      x_initial = x_wanted;
      y_initial = y_wanted;
      z_initial = z_wanted;
    }
  } else {
    // on est en mode progressif
    move_progressif(x_initial, x_wanted, y_initial, y_wanted, z_initial, z_wanted, iv);
  }
}

bool IVLoader::gotoposition(double x,double y,double z)
{
  // voici les variables de la premire mthode essaye.
  // elle correspond aux calculs rendus au rapport intermdiaire. Je suis quasiment sr que ces calculs contiennent une erreur quelque part...
  // Aprs des heures  essayer de corriger ces calculs, je me suis orient vers une autre solution...
  //double sin_beta,cos_beta,Z1,Z2,X,Y,sin_alpha,cos_alpha,A1,A2,sin_thetha,cos_thetha;

  double sin_2pi_sur_3 = 0.866025404f;
  bool en_dehors = false;

  double x1,C,l1,l2,alpha,beta,thetha;
  x1 = 0.2f;
  C = 0.043f;
  l1 = 0.34633f;
  l2 = 0.83f;
  /*
  // corps de calcul selon la premire mthode, pour le bras orient par (1,0,0)
  cos_beta = (z-Cz)/l2;
  sin_beta = sqrt(fabs(1-cos_beta*cos_beta));
  beta = atan2(sin_beta, cos_beta);

  Z1 = x- x1 + Cx;
  Z2 = -y;
  X = l1;
  Y = l2*sin_beta;
  cos_alpha = (Z1*Z1+Z2*Z2-X*X-Y*Y)/(2*X*Y);
  sin_alpha = -sqrt(fabs(1-cos_alpha*cos_alpha));
  A1 = X+(Y*cos_alpha);
  A2 = Y*(sin_alpha);
  sin_thetha = (A1*Z1-A2*Z2)/(A1*A1+A2*A2);
  cos_thetha = (A1*Z1+A2*Z2)/(A1*A1+A2*A2);
  thetha=atan2(sin_thetha, cos_thetha);

  alpha = atan2(sin_alpha, cos_alpha);
  */  
  /* 
  // je vous prsente galement une deuxime mthode brivement essaye.
  // ces calculs se basent sur les rsultats mathmatiques d'un camarade...
  // je n'ai pas vraiment persvr dans cette mthode, car je trouvais horribles les quations que cette mthode donne !
  angle_x = atan(x/ y) + acos((x* x+ y* y+ (z- a) * (z- a) + l1 * l1  - l2 *l2) / (l1 * sqrt( x* x+ y* y)));
  angle_y = atan((x- a*sqrt(3)/2) / y) + acos(((x- a*sqrt(3)/2) * (x- a*sqrt(3)/2) + y* y+ (z+ a/2) * (z+ a/2) + l1 * l1  - l2 *l2) / (l1 * sqrt( (x- a*sqrt(3)/2) * (x- a*sqrt(3)/2) + y* y)));
  angle_z = atan((x+ a*sqrt(3)/2) / y) + acos(((x+ a*sqrt(3)/2) * (x+ a*sqrt(3)/2) + y* y+ (z+ a/2) * (z+ a/2) + l1 * l1  - l2 *l2) / (l1 * sqrt( (x+ a*sqrt(3)/2) * (x+ a*sqrt(3)/2) + y* y)));
  */


  double x_wanted1, y_wanted1, z_wanted1;

  x_wanted1 = x+ C;
  y_wanted1 = y;
  z_wanted1 = z;


  // j'ai donc refais les calculs selon une (troisme !) mthode que je trouvais vraiment plus jolie.
  // pour le calcul, j'ai considr que le centre du plateau tait le centre de 3 sphres de rayon (l1+C). A la diffrence du calcul rendu au rapport intermdiaire, j'ai adopt pour une rsolution entirement gomtrique.
  double Qi,S,haut,bas;

  // Hormis les petites translations correspondant  la distance entre le centre du plateau et un des bords, la seule variable dont l'expression varie est Qi. En effet, Qi correspond au centre du plateau qui a subi une rotation, oppur se placer dans le mme repre que le bras numro i.
  // la forme gnrale de Qi est:
  // Qi = 2*x*cos(angle_i) - 2*z*sin(angle_i)
  // (le '-' vient de l'orientation choisie)
  Qi = 2*x_wanted1;

  S=(- x_wanted1*x_wanted1 - z_wanted1*z_wanted1 - y_wanted1*y_wanted1 + l2*l2 - l1*l1 - x1*x1)/l1;

  haut = (- 2*y_wanted1 - sqrt(4*(y_wanted1*y_wanted1+x1*x1) - S*S + Qi*Qi*(1-(x1*x1)/(l1*l1)) - ((2*Qi*x1*S)/l1) - (4*Qi*x1)));
  bas = - 2*x1 - S - Qi*((x1/l1)-1);

  thetha = 2*atan(haut/bas);
  if (thetha > 6.28318531)
    thetha -= 6.28318531;
  else if (thetha < 0)
    thetha += 6.28318531;

  beta = 2*asin(z_wanted1/(2*l2));

  if ( (l1*cos(thetha)+0.2) > x_wanted1){
    alpha = asin((-y_wanted1 - l1*sin(thetha))/(l2*cos(beta)));
  } else {
    alpha = 3.14159265 - asin((-y_wanted1 - l1*sin(thetha))/(l2*cos(beta)));
    en_dehors = true;
  }

  alpha = alpha + thetha; 
  if (alpha > 6.28318531)
    alpha -= 6.28318531;
  else if (alpha < 0)
    alpha += 6.28318531;

  double alpha1 = alpha;
  double beta1 = beta;
  double thetha1 = thetha;

  // expression pour la mthode 1
  //biceps1_rotor->rotation.setValue(SbVec3f(0, 0, 1),-2*1.57079633+thetha);
  //biceps1_rotor->rotation.setValue(SbVec3f(1, 0, 0),angle_x);
  //biceps1_rotor->rotation.setValue(SbVec3f(0, 1, 0),angle_y);
  //biceps1_rotor->rotation.setValue(SbVec3f(0, 0, 1),angle_z);
  // expression pour la mthode 2
  //avantbras1_rotor->rotation.setValue(SbVec3f(0, 0, 1),-3*1.57079633+alpha);
  //avantbras1_rotor2->rotation.setValue(SbVec3f(1, 0, 0),1.57079633-beta);
  //expression pour la mthode 3


  // de mme, pour le bras numro 2 (situ  2*Pi/3)
  double x_wanted2, y_wanted2, z_wanted2;

  x_wanted2 = x- C/2;
  y_wanted2 = y;
  z_wanted2 = z- C*sin_2pi_sur_3;

  Qi = - x_wanted2 - sin_2pi_sur_3*2*z_wanted2;

  S=(- x_wanted2*x_wanted2 - z_wanted2*z_wanted2 - y_wanted2*y_wanted2 + l2*l2 - l1*l1 - x1*x1)/l1;

  haut = (- 2*y_wanted2 - sqrt(4*(y_wanted2*y_wanted2+x1*x1) - S*S + Qi*Qi*(1-(x1*x1)/(l1*l1)) - ((2*Qi*x1*S)/l1) - (4*Qi*x1)));
  bas = - 2*x1 - S - Qi*((x1/l1)-1);

  thetha = 2*atan(haut/bas);
  if (thetha > 6.28318531)
    thetha -= 6.28318531;
  else if (thetha < 0)
    thetha += 6.28318531;

  beta = 2*asin((- z_wanted2/2 + sin_2pi_sur_3*x_wanted2)/(2*l2));

  if ( 2*0.866025404*(l1*cos(thetha)+0.2f) + x_wanted2*1.73205081 < - z_wanted2){
    alpha = 3.14159265 - asin((-y_wanted2 - l1*sin(thetha))/(l2*cos(beta)));
    en_dehors = true;
  } else {
    alpha = asin((-y_wanted2 - l1*sin(thetha))/(l2*cos(beta)));
  }

  if (alpha > 6.28318531)
    alpha -= 6.28318531;
  else if (alpha < 0)
    alpha += 6.28318531;
  alpha = alpha + thetha; 
  if (alpha > 6.28318531)
    alpha -= 6.28318531;
  else if (alpha < 0)
    alpha += 6.28318531;

  double alpha2 = alpha;
  double beta2 = beta;
  double thetha2 = thetha;


  // de mme, pour le bras numro 3 (situ  4*Pi/3)
  double x_wanted3, y_wanted3, z_wanted3;

  x_wanted3 = x- C/2;
  y_wanted3 = y;
  z_wanted3 = z+ C*sin_2pi_sur_3;

  Qi = - x_wanted3 + sin_2pi_sur_3*2*z_wanted3;

  S=(- x_wanted3*x_wanted3 - z_wanted3*z_wanted3 - y_wanted3*y_wanted3 + l2*l2 - l1*l1 - x1*x1)/l1;

  haut = (- 2*y_wanted3 - sqrt(4*(y_wanted3*y_wanted3+x1*x1) - S*S + Qi*Qi*(1-(x1*x1)/(l1*l1)) - ((2*Qi*x1*S)/l1) - (4*Qi*x1)));
  bas = - 2*x1 - S - Qi*((x1/l1)-1);

  thetha = 2*atan(haut/bas);
  if (thetha > 6.28318531)
    thetha -= 6.28318531;
  else if (thetha < 0)
    thetha += 6.28318531;

  beta = 2*asin((- z_wanted3/2 - sin_2pi_sur_3*x_wanted3)/(2*l2));

  if ( 2*0.866025404*(l1*cos(thetha)+0.2f) + x_wanted3*1.73205081 < z_wanted3){
    alpha = 3.14159265 - asin((-y_wanted3 - l1*sin(thetha))/(l2*cos(beta))) ;
    en_dehors = true;
  } else {
    alpha = asin((-y_wanted3 - l1*sin(thetha))/(l2*cos(beta))) ;
  }

  alpha = alpha + thetha; 
  if (alpha > 6.28318531)
    alpha -= 6.28318531;
  else if (alpha < 0)
    alpha += 6.28318531;

  double alpha3 = alpha;
  double beta3 = beta;
  double thetha3 = thetha;

  //faisons a prsent des tests pour savoir si les coordonnes son atteignables ou non.
  if (isnan(alpha1) || isnan(alpha2) || isnan(alpha3) || isnan(beta1) || isnan(beta2) || isnan(beta3) || isnan(thetha1) || isnan(thetha2) || isnan(thetha3) || y_wanted >= 0) 
    
    en_dehors = true;

  if (en_dehors) {
    char buffer[4151];
    sprintf(buffer,"Robot stopp  (%f,%f,%f), car (%f,%f,%f) est hors d'atteinte.", last_good_x, last_good_y, last_good_z, x_wanted, y_wanted, z_wanted);
    form1->textLabel1_4->setText(QString(buffer));
  } else {
    // deplaons le bras 1
    biceps1_rotor->rotation.setValue(SbVec3f(0, 0, -1),thetha1);
    avantbras1_rotor->rotation.setValue(SbVec3f(0, 0, -1),-(alpha1 - 1.57079633));
    avantbras1g_rotor2->rotation.setValue(SbVec3f(-1, 0, 0),beta1);
    avantbras1d_rotor2->rotation.setValue(SbVec3f(-1, 0, 0),beta1);
    avantbras1h_translation->translation= SbVec3f(0, -0.03*cos(beta1), 0.009+0.03*sin(beta1));
    avantbras1b_translation->translation= SbVec3f(0, -0.77*cos(beta1), 0.009+0.77*sin(beta1));
    // deplaons le bras 3
    biceps2_rotor->rotation.setValue(SbVec3f(0, 0, -1),thetha2);
    avantbras2_rotor->rotation.setValue(SbVec3f(0, 0, -1),-(alpha2 - 1.57079633));
    avantbras2g_rotor2->rotation.setValue(SbVec3f(-1, 0, 0),beta2);
    avantbras2d_rotor2->rotation.setValue(SbVec3f(-1, 0, 0),beta2);
    avantbras2h_translation->translation= SbVec3f(0, -0.03*cos(beta2), 0.009+0.03*sin(beta2));
    avantbras2b_translation->translation= SbVec3f(0, -0.77*cos(beta2), 0.009+0.77*sin(beta2));
    // deplaons le bras 3
    biceps3_rotor->rotation.setValue(SbVec3f(0, 0, -1),thetha3);
    avantbras3_rotor->rotation.setValue(SbVec3f(0, 0, -1),-(alpha3 - 1.57079633));
    avantbras3g_rotor2->rotation.setValue(SbVec3f(-1, 0, 0),beta3);
    avantbras3d_rotor2->rotation.setValue(SbVec3f(-1, 0, 0),beta3);
    avantbras3h_translation->translation= SbVec3f(0, -0.03*cos(beta3), 0.009+0.03*sin(beta3));
    avantbras3b_translation->translation= SbVec3f(0, -0.77*cos(beta3), 0.009+0.77*sin(beta3));
    // enfin, plaons le plateau par une simple translation.
    bottom_translation->translation= SbVec3f(x, y, z);
    return true;
  }
  return false;
}

IVLoader::~IVLoader()
{
  delete viewer;
}

void IVLoader::openFile(QString s)
{
  // Lecture du graphe de scne
  SoInput sceneInput;
  if ( !sceneInput.openFile(s.ascii()))
  { 
    QMessageBox::information((QWidget*)parent(), 
        "Erreur de lecture", 
        "Le fichier slectionn n'a pu tre ouvert.");
    return;
  }
  separator = SoDB::readAll(&sceneInput);
  if ( separator == NULL)
  { 
    QMessageBox::information((QWidget*)parent(), 
        "Erreur de lecture", 
        "Le fichier slectionn n'a pu tre dcod.");
    return ;
  }
  sceneInput.closeFile();

  // Association du graphe de scne au viewer
  viewer->setSceneGraph(separator);
  viewer->render();
}


void IVLoader::openMember(QString s, SoSeparator *new_separator, SoSeparator *root)
{
  // Lecture du graphe de scne
  SoInput sceneInput;
  if ( !sceneInput.openFile(s.ascii()))
  { 
    QMessageBox::information((QWidget*)parent(), 
        "Erreur de lecture", 
        "Le fichier slectionn n'a pu tre ouvert.");
    return;
  }

  new_separator = SoDB::readAll(&sceneInput);
  if ( new_separator == NULL)
  { 
    QMessageBox::information((QWidget*)parent(), 
        "Erreur de lecture", 
        "Le fichier slectionn n'a pu tre dcod.");
    return;
  }
  new_separator->ref();
  root->addChild(new_separator);
  sceneInput.closeFile();
  viewer->render();
}


void IVLoader::openFileDialog()
{
  // ouverture d'une bote de dialogue
  QString s = Q3FileDialog::getOpenFileName(
      QDir::currentDirPath(),
      "Inventor or VRML files (*.*)",
      (QWidget*)parent(),
      "open file dialog",
      "Choose a file to open" );

  // Si annulation alors arrter
  if (s.isEmpty())
    return ;

  openFile(s);
}

